from django.db import models
from logging import warn



class SerializerModel(models.Model):
    class Meta:
        abstract = True
        
    def convert(self,obj):
        # warn("【You should override this method】 convert %s" % obj,)
        return obj if isinstance(obj,str) else obj.__str__()
    
    def to_json(self, *args, **kwargs):
        return self.sample_to_json(*args, **kwargs)

    def loop_all_self_attr(
        self, with_foreign=True, with_related=False, related_serializer=False
    ):
        """遍历所有属性

        Returns:
            _type_: _description_
        """
        for key in self.__dict__:
            if hasattr(self, key):
                if key == "_state":
                    continue
                res = getattr(self, key)
                # test type res
                
                yield key, self.convert(res) if res is not None else None
        # print(self.foreignKeys())
        for fKey in self.foreignKeys():
            if hasattr(self, fKey.name) and with_foreign is True:
                foreign = getattr(self, fKey.name)
                # print("foreign", fKey.name, foreign)
                # one to one
                if hasattr(foreign, "sample_to_json"):
                    yield (
                        fKey.name,
                        getattr(self, fKey.name).sample_to_json(
                            with_foreign=True,
                            related_serializer=False,
                            with_related=True,
                        ),
                    )
                else:
                    # res = getattr(self, fKey.name)
                    # yield (
                    #     fKey.name,
                    #     res.__str__() if res is not None else None,
                    # )
                    # 可能是被别的对象引用或者 None ，被别的对象引用的话，这里是一个 RelatedManager 对象,不适合自动处理
                    yield (fKey.name, None)
            # related one to many
            if fKey.related_model is not None and with_related is True:
                related = fKey.related_model
                # print(self.__class__.__name__, fKey.name, related.__class__.__name__)
                arr = []
                if hasattr(related, self.__class__.__name__.lower()) is False:
                    continue

                for item in related.objects.filter(
                    **{self.__class__.__name__.lower(): self}
                ).all():
                    if item is not None and hasattr(item, "sample_to_json"):
                        if related_serializer is True:
                            arr.append(
                                item.sample_to_json(  # type: ignore
                                    with_foreign=False,
                                    related_serializer=False,
                                    with_related=False,
                                )  # type: ignore
                            )
                            continue
                        arr.append(item.__str__())
                yield fKey.name, arr
                yield fKey.name + "_count", len(arr)

    def extra_json(self):
        return {}

    def foreignKeys(self):
        return [f for f in self._meta.get_fields() if f.is_relation]

    def exclude_json_keys(self):
        return ["is_deleted", "password"]

    def sample_to_json(
        self,
        with_foreign=True,
        with_related=True,
        related_serializer=False,
        merge_force=False,
    ):
        """转换为json

        Returns:
            _type_: _description_
        """
        result = {
            key: value
            for key, value in self.loop_all_self_attr(
                with_foreign=with_foreign,
                with_related=with_related,
                related_serializer=related_serializer,
            )
        }

        for key, value in self.extra_json().items():
            if merge_force is True:
                result[key] = value
            else:
                if result.get(key) is None:
                    result[key] = value
                else:
                    warn("key %s is exists" % key)
        for key in self.exclude_json_keys():
            if result.keys().__contains__(key):
                del result[key]
        return result




from functools import wraps
import inspect
from django.http import HttpRequest
from django.urls import re_path
from backend import settings
from revolver_api.revolver_api.utils.get_request_args import  get_instance_from_args_or_kwargs
from revolver_api.revolver_api.response import ApiErrorCode, ApiJsonResponse


def valid_method_middlewares(method="GET"):
    def middleware(request:HttpRequest):
        print("valid_method_middlewares",request.method)
        if request.method == method:
            return True
        else:
            raise Exception("不支持的请求方法")
    return middleware


class Router():
    """_summary_
    """
    def __init__(self,baseUrl="api/") -> None:
        self.baseUrl = baseUrl
    
    routes = []
    routeMap = {}
    
    def route(self,url, middlewares=[],name_suffix="",
                exception_json=True,
                description=""
        ):
        """_summary_

        Args:
            url (_type_): _description_
            middlewares (list, optional): _description_. Defaults to [].
        """
        def decorator(func):
            
            @wraps(func)
            def inner(*args, **kwargs):
                request = get_instance_from_args_or_kwargs(HttpRequest, args, kwargs)
                for middleware in middlewares:
                    print("midd:",middleware.__name__)
                    try:
                        next = middleware(request)
                        if next:
                            return func(*args, **kwargs)
                    except Exception as e:
                        if settings.DEBUG:
                            raise e
                        if not exception_json:
                            raise e
                        return ApiJsonResponse.error(ApiErrorCode.ERROR,str(e))
                    
            # self.urls.append(path(url, inner, name=func.__name__ + name_suffix))
            file = inspect.getsourcefile(func)
            self.routeMap[url] = {
                "name": func.__name__ + name_suffix,
                "description": description,
                "file": "vscode://file/" + file + ":" + str(func.__code__.co_firstlineno),
                "func": inner,
            }
            self.routes.append({
                "url": "/" + url,
                "name": func.__name__ + name_suffix,
                "description": description,
                "file": "vscode://file/" + file + ":" + str(func.__code__.co_firstlineno),
            })
            return inner
            
        return decorator
    
    def get(self,url,middlewares=[],**kwargs):
        # TODO：合并 middlewares
        return self.route(url,middlewares=[valid_method_middlewares("GET"),*middlewares] ,name_suffix="_get",**kwargs)

    def post(self,url,middlewares=[],**kwargs):
        return self.route(url,middlewares=[valid_method_middlewares("POST")] + middlewares,name_suffix="_post",**kwargs)
    
    def put(self,url,middlewares=[],**kwargs):
        return self.route(url,middlewares=[valid_method_middlewares("PUT")] + middlewares,name_suffix="_put",**kwargs)
    
    def delete(self,url,middlewares=[],**kwargs):
        return self.route(url,middlewares=[valid_method_middlewares("DELETE")] + middlewares,name_suffix="_delete",**kwargs)
    
    def resource(self,baseUrl,middlewares=[],**kwargs):
        def wrapper(obj):
            o =  obj()
            if hasattr(o,"register"):
                print("register",baseUrl,middlewares)
                o.register(self,baseUrl,middlewares=middlewares,**kwargs)
        return wrapper
    
    @property
    def urls(self):
        return [
            # re_path(r"^api/(.*)$", Router.handler, name="api")
            re_path(r"^" + self.baseUrl + "(.*)$", Router.handler, name="api")
        ]
    
    
    @staticmethod
    def handler(request: HttpRequest,path: str):
        func = Router.routeMap.get(path)
        if func is None:
            return ApiJsonResponse.error(ApiErrorCode.ERROR,"未找到对应的路由")
        if path in ['users.detail1']:
            return ApiJsonResponse.error(ApiErrorCode.AUTH_ERROR,"未登录")
        return func["func"](request)
        
        
    






def get_instance_from_kwargs(cls, kwargs):
    """_summary_

    Args:
        cls (_type_): _description_
        kwargs (_type_): _description_

    Returns:
        _type_: _description_
    """
    for k, v in kwargs.items():
        if isinstance(v, cls):
            return v

def get_instance_from_args(cls, args):
    """_summary_

    Args:
        cls (_type_): _description_
        args (_type_): _description_

    Returns:
        _type_: _description_
    """
    for v in args:
        if isinstance(v, cls):
            return v
        
def get_instance_from_args_or_kwargs(cls, args, kwargs):
    """### 从参数或者关键字参数中获取实例

    Args:
        cls (_type_): _description_
        args (_type_): _description_
        kwargs (_type_): _description_

    Returns:
        _type_: _description_
    """
    return get_instance_from_args(cls, args) or get_instance_from_kwargs(cls, kwargs)


from pathlib import Path


def readTomlConfig(baseDir:Path):
    """_summary_

    Raises:
        Exception: _description_

    Returns:
        _type_: _description_
    """
    import toml

    configPath = str(baseDir)

    # if file exits
    if not Path(configPath).is_file():
        with open(configPath, "w") as f:
            f.write(
                '[timezone]\ntimezone = "Asia/Shanghai"\ndatetime_format = "%Y-%m-%d %H:%M:%S"'
            )
        raise Exception("config.toml not found,try to create a new one")

    with open(configPath, "r") as f:
        config = toml.load(f)
    return config


def get(config: dict[any],domain:str,default=""):  
    """_summary_

    Args:
        domain (_type_): _description_
        default (str, optional): _description_. Defaults to "".

    Returns:
        _type_: _description_
    """
    keys = domain.split(".")
    target = config
    for key in keys:
        if key in target:
            target = target[key]
        else:
            return default
    return target




import enum
from django.http import JsonResponse



            
            
class ApiErrorCode(enum.Enum):
    SUCCESS = 200, "成功"
    ERROR = 400, "失败"
    AUTH_ERROR = 401, "认证失败"
    AUTH_EXPIRED = 402, "认证过期"
    NOT_FOUND = 404, "资源不存在"
    
    USER_NOT_EXIST = 1001, "用户不存在"
    USER_EXIST = 1002, "用户已存在"
    USER_PASSWORD_ERROR = 1003, "密码错误"
    USER_PASSWORD_NOT_MATCH = 1004, "密码不匹配"
    
    USER_NOT_LOGIN = 1005, "用户未登录"
    USER_NOT_AUTH = 1006, "用户未认证"
    TOKEN_INVALID = 1007, "token无效"

class ApiJsonResponse(JsonResponse):
    def __init__(self, data, message="", code=ApiErrorCode.SUCCESS,httpCode=200, **kwargs):
        super().__init__(
            {
                "message": message or code.value[1], 
                "code": code.value[0], 
                "data": data
            },
            safe=False,
            json_dumps_params={"ensure_ascii": False, "indent": 4},
            **kwargs,
        )
        self["Access-Control-Allow-Origin"] = "*"
        self["Access-Control-Allow-Methods"] = "POST, GET, OPTIONS, PUT, DELETE"
        self["Access-Control-Max-Age"] = "1000"
        self["Access-Control-Allow-Headers"] = "*"
        self.status_code = httpCode
        
    def error(code=ApiErrorCode.ERROR,message="错误",data=None,**kwargs):
        return ApiJsonResponse(data,code=code,message=message,httpCode=400,**kwargs)
    
    def success(data=None,message="成功",**kwargs):
        return ApiJsonResponse(data,code=ApiErrorCode.SUCCESS,message=message,httpCode=200,**kwargs)




from functools import wraps
import re
from typing import Any, Iterable
from django.http import HttpRequest, JsonResponse
from backend import settings
from revolver_api.revolver_api.utils.get_request_args import get_instance_from_args_or_kwargs
from core.models import BaseModel
from revolver_api.revolver_api.response import ApiErrorCode, ApiJsonResponse
from revolver_api.revolver_api.route import Router




def errorHandler(json=True):
    """api 错误处理

    Args:
        json (bool, optional): _description_. Defaults to True.
    """
    def wrapper(func):
        print("errorHandler",func.__name__)
        
        @wraps(func)
        def err_inner(*args, **kwargs):
            print("errorHandler inner")
            try:
                print("working fine....")
                return func(*args, **kwargs)
            except Exception as e:
                if json is False:
                    raise e
                return ApiJsonResponse({} if not hasattr(e,'data') else getattr(e,'data'), code=ApiErrorCode.ERROR,message=str(e) or "error")
        return err_inner
    return wrapper



class ApiException(Exception):
    code = 400
    message = "api exception"


class Rule:
    name: str = ""
    required: bool = False
    type: str = "string"
    max_length: int = 0
    min_length: int = 0
    max: int = 0
    min: int = 0
    choices: Iterable = []
    default: str = ""
    message: str = ""
    validator = bool = lambda *args, **kwargs: True

    def __init__(self,name, message="", **kwargs):
        self.name = name
        self.message = message
        for key in kwargs:
            setattr(self, key, kwargs[key])
            
    def is_required(self):
        self.required = True
        return self
    def string(self):
        self.type = "string"
        return self
    def number(self):
        self.type = "number"
        return self
    def set_choices(self,choices):
        self.choices = choices
        return self
    def set_min(self,min):
        self.min = min
        return self
    def set_max(self,max):
        self.max = max
        return self
    def set_message(self,message):
        self.message = message
        return self
    

def validator(rules: Iterable[Rule]=[],method="get"):
    def wrapper(func):
        @wraps(func)
        def inner(*args,**kwargs):
            req = get_instance_from_args_or_kwargs(HttpRequest,args,kwargs)
            params = {}
            if method.lower() == 'get':
                params = req.GET.dict()
            else:
                params = req.POST.dict()
            
            # print("!!! params:",params)
            for rule in rules:
                value = params.get(rule.name)
                if rule.required and value is None or value == "":
                    return ApiJsonResponse(None,code=ApiErrorCode.ERROR,message=rule.message)
                
            return func(*args,**kwargs)
        return inner
    return wrapper
class Api:
    """# 生成API

    Returns:
        _type_: _description_
    """

    model: BaseModel

    rules: Iterable[Rule] = []


    class Validator:
        is_valid = True
        errors = {}

        @property
        def tips(self):
            for key in self.errors:
                return self.errors[key]

        def add_error(self, key, value):
            self.is_valid = False
            self.errors[key] = value

    def validate(self, request: HttpRequest, **kwargs):
        """### 提交数据验证

        Args:
            request (HttpRequest): _description_

        Returns:
            _type_: _description_
        """
        print(self.model, "validate")
        validator = self.Validator()
        for rule in self.rules:
            value = request.POST.get(rule.name)
            if rule.required is True and (value is None or value == ""):
                validator.add_error(
                    rule.name, rule.message if rule.message else "required"
                )
                continue
            if rule.type == "string":
                value = value if value else ""
                if rule.max_length > 0 and len(value) > rule.max_length:
                    validator.add_error(rule.name, f"max_length {rule.max_length}")
                    continue
                if rule.min_length > 0 and len(value) < rule.min_length:
                    validator.add_error(rule.name, f"min_length {rule.min_length}")
                    continue
            if rule.type == "number":
                value = int(value if value else 0)
                if rule.max > 0 and value > rule.max:
                    validator.add_error(rule.name, f"max {rule.max}")
                    continue
                if rule.min > 0 and value < rule.min:
                    validator.add_error(rule.name, f"min {rule.min}")
                    continue
            if rule.type == "choices":
                if value not in rule.choices:
                    validator.add_error(rule.name, f"choices is not in {rule.choices}")
                    continue
            if rule.validator(value) is False:
                validator.add_error(
                    rule.name, rule.message if rule.message else "自定义验证错误"
                )
                continue
        return validator

    def create(self, request: HttpRequest, **kwargs):
        """### 创建数据

        Args:
            request (HttpRequest): _description_

        Returns:
            _type_: _description_
        """
        if request.method != "POST":
            return JsonResponse({"error": "only support POST"})
        print(self.model, "createApi")
        validator = self.validate(request, **kwargs)
        if validator.is_valid is False:
            return JsonResponse(
                {
                    "code": 400,
                    "msg": "validate error",
                    "tips": validator.tips if validator.tips else "",
                    "errors": validator.errors,
                }
            )
        try:
            dict = request.POST.dict()
            del dict["id"]
            obj = self.model.objects.create(**dict)
        except Exception as e:
            return JsonResponse({"error": str(e)})
        return JsonResponse(
            {
                "status": "success",
                "code": 200,
                "data": obj.to_json(),
            }
        )

    def update(self, request: HttpRequest, **kwargs):
        """### 更新数据

        Args:
            request (HttpRequest): _description_

        Returns:
            _type_: _description_
        """
        if request.method != "POST":
            return JsonResponse({"error": "only support POST"})
        # print(self.model, "createApi")
        validator = self.validate(request, **kwargs)
        if validator.is_valid is False:
            return JsonResponse(
                {
                    "code": 400,
                    "msg": "validate error",
                    "tips": validator.tips if validator.tips else "",
                    "errors": validator.errors,
                }
            )
        id = request.POST.get("id")
        if id is None or id == "":
            return JsonResponse({"error": "id is required"})
        obj = self.model.objects.filter(pk=id).first()
        if obj is None:
            return JsonResponse({"error": "not found"})
        for key in request.POST.dict():
            setattr(obj, key, request.POST.dict()[key])
        try:
            obj.save()
        except Exception as e:
            return JsonResponse({"error": str(e)})
        return JsonResponse(
            {
                "status": "success",
                "code": 200,
                "msg": "update success",
                "data": obj.to_json(),
            }
        )
        
    def defaultQuery(self,request: HttpRequest):
        """### 默认查询

        Args:
            request (HttpRequest): _description_

        Returns:
            _type_: _description_
        """
        
        model_fields = self.model._meta.get_fields()
        valid_fields = {}
        input_fields = request.GET.dict()
        
        # 忽略不在模型中的字段，前端在查询时可能会传入一些不在模型中的字段，这些字段应该被忽略
        for field in model_fields:
            for key in input_fields:
                if not re.match(f'{field.name}__\w+',key):
                    pass
                else:
                    valid_fields[key] = input_fields[key]
        return self.model.objects.all().filter(**valid_fields).order_by("-created_at")
        
    def list(self, request: HttpRequest, **kwargs):
        if request.method != "GET":
            return JsonResponse({"error": "only support GET"})
        print(self.model, "pageApi")
        if self.validate(request, **kwargs) is False:
            return JsonResponse({"error": "validate error"})
        page, size = request.GET.get("page", 1), request.GET.get("size", 10)
        page = int(page)
        size = int(size)
        objs = self.defaultQuery(request=request)
        count = objs.count()
        objs = objs[
            (page - 1) * size : (page) * size
        ]
        arr = []
        for obj in objs:
            if hasattr(obj, "to_json"):
                arr.append(obj.to_json())
            else:
                arr.append(obj)
        return ApiJsonResponse(
            {
                "pageable": {
                    "page": page,
                    "size": size,
                    "total": count,
                    "totalPage": count // size + 1,
                },
                "list": arr,
            },
            message="获取成功",
        )

    @validator([
        Rule(name="id", required=True, message="id不能为空"),
    ])
    def detail(self, request: HttpRequest):
        if request.method != "GET":
            return JsonResponse({"error": "only support GET"})
        id = request.GET.get("id")
        # print(self.model, "get_one",id)
        try:
            obj = self.model.objects.get(id=id)
        except Exception as e:
            if settings.DEBUG:
                raise e
            return ApiJsonResponse.error(ApiErrorCode.NOT_FOUND,"没有 id 为 "+ id +" 的找到记录")
        return JsonResponse(
            {
                "status": "success",
                "code": 200,
                "data": obj.to_json(),
            }
        )

    def delete(self, request: HttpRequest):
        id = request.GET.get("id")
        if request.method != "DELETE":
            return JsonResponse({"error": "only support DELETE"})
        print(self.model, "deleteApi")
        obj = self.model.objects.get(id=id)
        if obj is None:
            return JsonResponse({"error": "not found"})
        obj.delete()
        return JsonResponse(
            {
                "status": "success",
                "code": 200,
                "data": obj.to_json(),
            }
        )

    @property
    def urls(self):
        return self.get_urls(), "api", self.model.__name__.lower()
    
    @property
    def routeName(self):
        return self.model.__name__.lower()

    def __init__(self, *args: Any, **kwds: Any) -> Any:
        pass
    
    def register(self,router:Router,baseUrl="api",middlewares=[]):
        router.get(baseUrl,middlewares=middlewares)(self.list)
        router.post(baseUrl + '.create',middlewares=middlewares)(self.create)
        router.get(baseUrl + '.detail',middlewares=middlewares)(self.detail)
        router.delete(baseUrl + '.delete',middlewares=middlewares)(self.delete)
        router.put(baseUrl + '.update',middlewares=middlewares)(self.update)
    
